#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <time.h>
#include <errno.h>

#include "ilgate.h"

unsigned long 
microtime ()
{
	struct timeval tv;

	gettimeofday (&tv, NULL);
	return (tv.tv_sec * 1000000 + tv.tv_usec);
}

int in_cksum (void *buf, int n);

enum ilstate {
	ils_closed,
	ils_syncer,
	ils_syncee,
	ils_established,
	ils_closing,
};

char *ilstatenames[] = {
	"closed ",
	"syncer ",
	"syncee ",
	"estab  ",
	"closing",
};

enum iltype {
	ilt_sync = 0,
	ilt_data = 1,
	ilt_dataquery = 2,
	ilt_ack = 3,
	ilt_query = 4,
	ilt_state = 5,
	ilt_close = 6,
};

char *iltypenames[] = {
	"sync",
	"data",
	"dataquery",
	"ack",
	"query",
	"state",
	"close",
};

struct ilconn {
	struct ilconn *lnext;
	struct ilconn *lprev;
	struct in_addr raddr;
	int rport;
	struct in_addr laddr;
	int lport;
	enum ilstate state;
	int connid;
	unsigned long rid0;
	unsigned long id0;
	unsigned long next;
	unsigned long rcvd;
	int (*func)(void *, void *, int);
	int srv_state;
	struct ilbuf *rcv, *rcv_tail;
	struct ilbuf *snd, *snd_tail;
	unsigned long last_xmit; /* usec */
	int rtt; /* usec */
	int retransmit_mode;
	unsigned long need_ack;
	int retransmit_factor;
	unsigned long next_rexmit_time;
	struct ilbuf *future;
};

struct ilbuf {
	struct ilbuf *next;
	int len;
	int id;
	unsigned long timestamp;
	char data[0];
};

void il_sendctl (struct ilconn *ilconn, enum iltype type, int id, int ack);
void il_send (struct ilconn *ilconn, struct ilbuf *ilbuf, enum iltype type);
int il_rcv_request (struct ilconn *ilconn);

void
dump (void *buf, int len)
{
	unsigned char *p;
	int off;
	int i;
	int c;

	p = buf;
	for (off = 0; off < len; off += 16) {
		printf ("%04x ", off);

		for (i = 0; i < 16; i++) {
			if (i == 8)
				putchar (' ');
			if (off + i < len)
				printf ("%02x ", p[off+i]);
			else
				printf ("   ");
		}

		putchar (' ');
		for (i = 0; i < 16; i++) {
			if (off + i < len) {
				c = p[off+i];
				if (c >= ' ' && c < 0177)
					putchar (c);
				else
					putchar ('.');
			} else {
				putchar (' ');
			}
		}
		putchar ('\n');
	}
}

struct il {
	unsigned char il_sum[2];
	unsigned char il_len[2];
	unsigned char il_type;
	unsigned char il_spec;
	unsigned char il_sport[2];
	unsigned char il_dport[2];
	unsigned char il_id[4];
	unsigned char il_ack[4];
};

int
il_getshort (unsigned char *p)
{
	return ((p[0] << 8) | p[1]);
}

int
il_getlong (unsigned char *p)
{
	return ((p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3]);
}

void
il_putshort (unsigned char *p, int val)
{
	p[0] = val >> 8;
	p[1] = val;
}

void
il_putlong (unsigned char *p, int val)
{
	p[0] = val >> 24;
	p[1] = val >> 16;
	p[2] = val >> 8;
	p[3] = val;
}

char *
il_statename (enum ilstate state)
{
	if ((unsigned)state < sizeof ilstatenames / sizeof ilstatenames[0])
		return (ilstatenames[state]);
	else
		return ("bad");
}

char *
il_typename (enum iltype type)
{
	if ((unsigned)type < sizeof iltypenames / sizeof iltypenames[0])
		return (iltypenames[type]);
	else
		return ("bad");
}


int ilconnid;
struct ilconn ilconns;

struct ilconn *
find_ilconn (struct in_addr src, struct in_addr dst, struct il *ilp)
{
	int sport, dport;
	struct ilconn *ilconn;

	if (ilconns.lnext == NULL) {
		ilconns.lnext = &ilconns;
		ilconns.lprev = &ilconns;
	}

	sport = il_getshort (ilp->il_sport);
	dport = il_getshort (ilp->il_dport);

	for (ilconn = ilconns.lnext;
	     ilconn != &ilconns;
	     ilconn = ilconn->lnext)
		if (ilconn->raddr.s_addr == src.s_addr
		    && ilconn->rport == sport
		    && ilconn->laddr.s_addr == dst.s_addr
		    && ilconn->lport == dport)
			return (ilconn);

	if (ilp->il_type == ilt_close)
		return (NULL);

	ilconn = calloc (1, sizeof *ilconn);

	ilconn->raddr = src;
	ilconn->rport = sport;
	ilconn->laddr = dst;
	ilconn->lport = dport;
	ilconn->state = ils_closed;
	ilconn->connid = ++ilconnid;

	ilconn->lnext = &ilconns;
	ilconn->lprev = ilconns.lprev;
	ilconns.lprev->lnext = ilconn;
	ilconns.lprev = ilconn;

	return (ilconn);
}

void
il_delete (struct ilconn *ilconn)
{
	struct ilbuf *ilbuf;

	ilconn->lnext->lprev = ilconn->lprev;
	ilconn->lprev->lnext = ilconn->lnext;
	while ((ilbuf = ilconn->rcv) != NULL) {
		ilconn->rcv = ilbuf->next;
		free (ilbuf);
	}
	while ((ilbuf = ilconn->snd) != NULL) {
		ilconn->snd = ilbuf->next;
		free (ilbuf);
	}
	while ((ilbuf = ilconn->future) != NULL) {
		ilconn->future = ilbuf->next;
		free (ilbuf);
	}
	free (ilconn);
}

void
il_close (struct ilconn *ilconn)
{
	if (ilconn->state != ils_closing) {
		ilconn->state = ils_closing;
		il_sendctl (ilconn, ilt_close, ilconn->next, ilconn->rcvd);
	}
}

void
il_abort (struct ilconn *ilconn)
{
	il_close (ilconn);
}

void
ilprintf (struct ilconn *ilconn, char *fmt, ...)
{
	va_list args;

	printf ("il%d %s: ", ilconn->connid, il_statename (ilconn->state));
	va_start (args, fmt);
	vprintf (fmt, args);
	va_end (args);
}

void
ilrcv (struct in_addr src, struct in_addr dst, void *pkt, int pktlen)
{
	struct il *il;
	struct ilconn *ilconn;
	unsigned long id, ack;
	int illen;
	void *data;
	int datalen;
	struct ilbuf *ilbuf;

	if (pktlen < sizeof *il) {
		printf ("il pkt too small\n");
		return;
	}

	il = pkt;

	illen = il_getshort (il->il_len);

	if (illen < sizeof *il) {
		printf ("bad illen\n");
		return;
	}

	if (illen > pktlen) {
		printf ("pkt truncated\n");
		return;
	}

	if (in_cksum (il, illen) != 0) {
		printf ("il cksum error\n");
		return;
	}

	if ((ilconn = find_ilconn (src, dst, il)) == NULL)
		return;

	id = il_getlong (il->il_id);
	ack = il_getlong (il->il_ack);
	data = &il[1];
	datalen = illen - sizeof *il;

	if (debug_flag)
		ilprintf (ilconn, "rcv %s(%d,%d) %d\n",
			  il_typename (il->il_type),
			  id - ilconn->rid0,
			  ack - ilconn->id0,
			  datalen);

	if (ack >= ilconn->id0 && ack <= ilconn->next) {
		while ((ilbuf = ilconn->snd) != NULL) {
			if (ack < ilbuf->id)
				break;
			ilconn->snd = ilbuf->next;
			free (ilbuf);
		}

		if (ilconn->snd == NULL)
			ilconn->retransmit_mode = 0;
	}

	switch (ilconn->state) {
	case ils_closed:
		switch (il->il_type) {
		case ilt_sync:
			if (debug_flag)
				ilprintf (ilconn, "new state syncee %d->%d\n",
					  ilconn->rport, ilconn->lport);
			ilconn->state = ils_syncee;
			ilconn->rid0 = il_getlong (il->il_id);
			ilconn->id0 = random () & 0x00ffffff;

			if (debug_flag)
				ilprintf (ilconn, "rid0 %d id0 %d\n",
					  ilconn->rid0, ilconn->id0);

			ilconn->rcvd = ilconn->rid0;
			ilconn->next = ilconn->id0 + 1;

			if (il_rcv_request (ilconn) < 0)
				goto done;
			goto syncee;

		case ilt_data: 
		case ilt_dataquery: 
		case ilt_ack: 
		case ilt_query: 
		case ilt_state: 
			ilconn->rcvd = id;
			ilconn->next = 0;
			il_abort (ilconn);
			goto done;

		case ilt_close: 
			il_delete (ilconn);
			goto done;
		}
		ilprintf (ilconn, "invalid pkt type: %d\n", il->il_type);
		il_abort (ilconn);
		goto done;

	case ils_syncer:
		switch (il->il_type) {
		case ilt_sync: break;
		case ilt_data: break;
		case ilt_dataquery: break;
		case ilt_ack: break;
		case ilt_query: break;
		case ilt_state: break;
		case ilt_close: break;
		}
		ilprintf (ilconn, "invalid pkt type %d\n", il->il_type);
		il_abort (ilconn);
		goto done;

	case ils_syncee:
	syncee:
		switch (il->il_type) {
		case ilt_sync:
			if (id != ilconn->rid0) {
				il_delete (ilconn);
				goto done;
			}

			if (ack != 0) {
				il_delete (ilconn);
				goto done;
			}

			il_sendctl (ilconn, ilt_sync,
				    ilconn->id0, ilconn->rid0);
			goto done;

		case ilt_data: 
		case ilt_dataquery: 
			goto done;

		case ilt_ack: 
			/* XXX */
			if (ack != ilconn->id0)
				goto done;
			if (debug_flag)
				ilprintf (ilconn, "new state estab\n");
			ilconn->state = ils_established;
			goto done;

		case ilt_query: 
		case ilt_state: 
			goto done;

		case ilt_close:
			if (ack == ilconn->id0)
				il_delete (ilconn);
			goto done;
		}
		ilprintf (ilconn, "invalid pkt type %d\n", il->il_type);
		il_abort (ilconn);
		goto done;

	case ils_established:
		switch (il->il_type) {
		case ilt_sync: 
			goto done;

		case ilt_data: 
		case ilt_dataquery:
			if (id <= ilconn->rcvd)
				goto done;

			if (id > ilconn->rcvd + 1 && id < ilconn->rcvd + 10) {
				for (ilbuf = ilconn->future;
				     ilbuf;
				     ilbuf = ilbuf->next)
					if (ilbuf->id == id)
						goto done;
			}
			
			ilbuf = malloc (sizeof *ilbuf + datalen);
			memset (ilbuf, 0, sizeof *ilbuf);
			ilbuf->len = datalen;
			ilbuf->id = id;
			memcpy (ilbuf->data, data, datalen);
			ilbuf->next = NULL;

			if (id != ilconn->rcvd + 1) {
				struct ilbuf **pp;
				pp = &ilconn->future;
				while (*pp && (*pp)->id < id)
					pp = &(*pp)->next;
				ilbuf->next = *pp;
				*pp = ilbuf;
				goto done;
			}

			ilconn->rcvd = id;
			if (ilconn->rcv == NULL)
				ilconn->rcv = ilbuf;
			else
				ilconn->rcv_tail->next = ilbuf;
			ilconn->rcv_tail = ilbuf;

			while ((ilbuf = ilconn->future) != NULL
			       && ilbuf->id == ilconn->rcvd + 1) {
				ilconn->future = ilbuf->next;
				ilconn->rcvd = ilbuf->id;
				ilconn->rcv_tail->next = ilbuf;
				ilconn->rcv_tail = ilbuf;
			}

			if (il->il_type == ilt_data)
				ilconn->need_ack = microtime ();
			else /* dataquery */
				il_sendctl (ilconn, ilt_state,
					    ilconn->next, ilconn->rcvd);
			goto done;

		case ilt_ack: 
			goto done;

		case ilt_query: 
			il_sendctl (ilconn, ilt_state,
				    ilconn->next, ilconn->rcvd);
			goto done;

		case ilt_state: 
			if ((ilbuf = ilconn->snd) != NULL)
				il_send (ilconn, ilbuf, ilt_dataquery);
			goto done;

		case ilt_close: 
			if (id == ilconn->rcvd + 1)
				il_abort (ilconn);
			goto done;
		}
		ilprintf (ilconn, "invalid pkt type %d\n", il->il_type);
		il_abort (ilconn);
		goto done;

	case ils_closing:
		switch (il->il_type) {
		case ilt_sync: 
		case ilt_data:
		case ilt_dataquery: 
		case ilt_ack:
		case ilt_query:
		case ilt_state:
			goto done;

		case ilt_close: 
			if (ack == ilconn->next) {
				ilconn->rcvd = id;
				il_sendctl (ilconn, ilt_close,
					    ilconn->next, ilconn->rcvd);
				il_delete (ilconn);
			}
			goto done;
		}
		ilprintf (ilconn, "invalid pkt type %d\n", il->il_type);
		goto done;
	}
		
	printf ("il state machine error\n");
	il_abort (ilconn);

 done:
	if ((ilbuf = ilconn->future) != NULL
	    && ilbuf->id == ilconn->rcvd + 1) {
	}

	return;
}

void
il_maybe_rexmit (struct ilconn *ilconn)
{
	struct ilbuf *ilbuf;
	unsigned long now;

	if ((ilbuf = ilconn->snd) == NULL)
		return;

	now = microtime ();

	if (ilconn->retransmit_mode == 0) {
		if (now - ilbuf->timestamp < ilconn->rtt * 4)
			return;
		ilconn->retransmit_mode = 1;
		ilconn->retransmit_factor = 2;
	} else if (now < ilconn->next_rexmit_time) {
		return;
	}

	ilconn->retransmit_factor 
		= ilconn->retransmit_factor * 3 / 2;
	il_send (ilconn, ilbuf, ilt_dataquery);
	ilconn->next_rexmit_time
		= now + ilconn->retransmit_factor * ilconn->rtt;
}

/* called every 100ms */
void
il_timer ()
{
	struct ilconn *ilconn, *nextil;

	for (ilconn = ilconns.lnext, nextil = ilconn->lnext;
	     ilconn != &ilconns; 
	     ilconn = nextil, nextil = ilconn->lnext) {
		il_maybe_rexmit (ilconn);
		if (ilconn->need_ack
		    && microtime () - ilconn->need_ack >= 200000)
			il_sendctl (ilconn, ilt_sync,
				    ilconn->next, ilconn->rcvd);
	}
}

void
il_sendctl (struct ilconn *ilconn, enum iltype type, int id, int ack)
{
	char buf[1000];
	struct il *il;
	struct sockaddr_in addr;
	int cksum;

	ilconn->need_ack = 0;

	il = (struct il *)buf;

	memset (il, 0, sizeof *il);
	il_putshort (il->il_len, sizeof *il);
	il->il_type = type;
	il_putshort (il->il_sport, ilconn->lport);
	il_putshort (il->il_dport, ilconn->rport);
	il_putlong (il->il_id, id);
	il_putlong (il->il_ack, ack);

	cksum = in_cksum (il, sizeof *il);
	il->il_sum[0] = cksum;
	il->il_sum[1] = cksum >> 8;

	memset (&addr, 0, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_addr = ilconn->raddr;
	addr.sin_port = 0;

	if (debug_flag)
		ilprintf (ilconn, "snd %s(%d,%d)\n",
			  il_typename (il->il_type),
			  id - ilconn->id0,
			  ack - ilconn->rid0);

	if (sendto (raw_sock, buf, sizeof *il,
		    0, (struct sockaddr *)&addr, sizeof addr) < 0)
		printf ("sendto error: %s\n", strerror (errno));
}

void
il_send (struct ilconn *ilconn, struct ilbuf *ilbuf, enum iltype type)
{
	char buf[10000];
	struct il *il;
	struct sockaddr_in addr;
	int cksum;

	ilconn->need_ack = 0;

	if (sizeof (struct il) + ilbuf->len > sizeof buf) {
		il_abort (ilconn);
		return;
	}

	il = (struct il *)buf;

	memset (il, 0, sizeof *il);
	il_putshort (il->il_len, sizeof *il + ilbuf->len);
	il->il_type = type;
	il_putshort (il->il_sport, ilconn->lport);
	il_putshort (il->il_dport, ilconn->rport);
	il_putlong (il->il_id, ilbuf->id);
	il_putlong (il->il_ack, ilconn->rcvd);

	memcpy (&il[1], ilbuf->data, ilbuf->len);

	cksum = in_cksum (il, sizeof *il + ilbuf->len);
	il->il_sum[0] = cksum;
	il->il_sum[1] = cksum >> 8;

	memset (&addr, 0, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_addr = ilconn->raddr;
	addr.sin_port = 0;

	if (debug_flag)
		ilprintf (ilconn, "send %s(%d,%d)\n",
			  il_typename (il->il_type),
			  ilconn->next - ilconn->id0,
			  ilconn->rcvd - ilconn->rid0);

	if (sendto (raw_sock, buf, sizeof *il + ilbuf->len,
		    0, (struct sockaddr *)&addr, sizeof addr) < 0)
		printf ("sendto error: %s\n", strerror (errno));

}

int
il_rcv_request (struct ilconn *ilconn)
{
	switch (ilconn->lport) {
	case 566:
		make_task (auth_srv, ilconn);
		return (0);
	default:
		ilprintf (ilconn, "no server for port %d\n", ilconn->lport);
		il_abort (ilconn);
		return (-1);
	}
}

int
il_read (struct ilconn *ilconn, char *buf, int len)
{
	struct ilbuf *ilbuf;
	int thistime;

	if ((ilbuf = ilconn->rcv) == NULL)
		return (0);

	thistime = ilbuf->len;
	if (len < thistime)
		thistime = len;

	memcpy (buf, ilbuf->data, thistime);

	ilconn->rcv = ilbuf->next;
	free (ilbuf);

	return (thistime);
}

int
il_write (struct ilconn *ilconn, char *buf, int len)
{
	struct ilbuf *ilbuf;

	ilbuf = malloc (sizeof *ilbuf + len);
	memset (ilbuf, 0, sizeof ilbuf);
	ilbuf->len = len;
	memcpy (ilbuf->data, buf, len);
	ilbuf->id = ilconn->next++;
	ilbuf->timestamp = microtime ();

	ilbuf->next = NULL;
	if (ilconn->snd == NULL)
		ilconn->snd = ilbuf;
	else
		ilconn->snd_tail->next = ilbuf;
	ilconn->snd_tail = ilbuf;

	if (ilconn->retransmit_mode == 0)
		il_send (ilconn, ilbuf, ilt_data);

	return (len);
}

